""" A multi function mapping UFL expressions to pymbolic expressions

This mapper only defines handler for those UFL expression types that have an
equivalent in pymbolic (mainly algebraic operations).

This is mainly intended as a base class for anything mapping ufl to pymbolic.
"""

from __future__ import absolute_import
from ufl.algorithms import MultiFunction
from pymbolic.primitives import Product, Quotient, Subscript, Sum, Variable

# The constructed pymbolic expressions use n-ary operators instead of ufls binary operators
from dune.perftool.ufl.flatoperators import get_operands


class UFL2PymbolicMapper(MultiFunction):
    def __init__(self):
        super(UFL2PymbolicMapper, self).__init__()

    call = MultiFunction.__call__

    def product(self, o):
        return Product(tuple(self.call(op) for op in get_operands(o)))

    def multi_index(self, o):
        return tuple(self.call(op) for op in o.ufl_operands)

    def index(self, o):
        # One might as well take the uflname as string here, but I apply this function
        from dune.perftool.pdelab import name_index

        return Variable(name_index(o))

    def fixed_index(self, o):
        return o._value

    def indexed(self, o):
        aggr = self.call(o.ufl_operands[0])
        ind = self.call(o.ufl_operands[1])

        # If the left operand is already a subscripted pymbolic expression, we need to do tuple addition on the indices
        if isinstance(aggr, Subscript):
            return Subscript(aggr.aggregate, aggr.index + ind)
        else:
            return Subscript(aggr, ind)

    def float_value(self, o):
        return o.value()

    def int_value(self, o):
        return o.value()

    def division(self, o):
        assert len(o.ufl_operands) == 2

        return Quotient(self.call(o.ufl_operands[0]), self.call(o.ufl_operands[1]))

    def sum(self, o):
        return Sum(tuple(self.call(op) for op in get_operands(o)))
